<?php

define('ROOT_PATH', '/usr/share/phpmyadmin/');
define('TEST_PATH', __DIR__ . '/../../');
define('AUTOLOAD_FILE', __DIR__ . '/../../test/autoload.php');
require_once(__DIR__ . '/../../test/bootstrap-dist.php');
